// SaslSspi.cpp : Defines the entry point for the application.
//

#include <shlwapi.h>
#include "net_za_darkskies_security_sasl_SSPIImpl.h"
#include "precomp.h"
#include "model.h"

static jclass clsSSPIImpl;
static jfieldID fidModel;
static jfieldID fidAuthorizationID;
static jfieldID fidCompleted;
static jmethodID midCompleteAuthorization;

JNIEXPORT void JNICALL Java_net_za_darkskies_security_sasl_SSPIImpl_initIDs
  (JNIEnv * env, jclass jcls)
{
	clsSSPIImpl = jcls;
	fidModel = env->GetFieldID(jcls, "model", "J");
	fidAuthorizationID = env->GetFieldID(jcls, "authorizationID", "Ljava/lang/String;");
	fidCompleted = env->GetFieldID(jcls, "completed", "Z");
	midCompleteAuthorization = env->GetMethodID(jcls, "completeAuthorization", "(Ljava/lang/String;Ljava/lang/String;)V");
}

void _err(JNIEnv * env, const wchar_t* psz, bool bRecoverable = false, DWORD err = GetLastError()) {

	wchar_t szErr[256];
	if (!FormatMessageW(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
		0, err, 0, szErr, sizeof szErr / sizeof *szErr, 0))
		wnsprintfW(szErr, 255, L"Unknown Error: %x", err);

	wchar_t szMsg[512];
	wnsprintfW(szMsg, 511, L"%s failed (0x%08x): %s", psz, err, szErr);
	szMsg[511] = 0;

	char szBuffer[512];
	WideCharToMultiByte( CP_UTF8, 0, szMsg, -1, szBuffer, 512, NULL, NULL );

	size_t eos = strlen(szBuffer) - 1;
	if (szBuffer[eos-1] = '\r') szBuffer[eos-1] = 0;
	else if (szBuffer[eos] = '\n') szBuffer[eos] = 0;

#ifdef _DEBUG
	wprintf(L"Exception: %s\n", szMsg);
#endif

	jclass clsExc = env->FindClass("javax/security/sasl/SaslException");
	jmethodID midExc = env->GetMethodID(clsExc, "<init>", "(Ljava/lang/String;)V");

	jstring str = env->NewStringUTF(szBuffer);
	jthrowable e2 = (jthrowable) env->NewObject(clsExc, midExc, str);
	env->Throw(e2);

	env->DeleteLocalRef(clsExc);

	//env->ThrowNew(clsExc, szBuffer);
#ifdef _DEBUG
	wprintf(L"Exception thrown\n");
#endif
	//__asm int 3;
}

jobjectArray getMechanismNames
  (JNIEnv *env, jclass jobj, bool isServer, jobject map);

JNIEXPORT void JNICALL Java_net_za_darkskies_security_sasl_SSPIImpl_clientInit
  (JNIEnv *env, jobject jobj, jstring mechanisms, jstring protocol, jstring serverName, jobject props, jobject cbh)
{

	Model* s_pModel = new Model(env, false, _err);
	env->SetLongField(jobj, fidModel, (long long)s_pModel);

	const jchar *mechs = env->GetStringChars(mechanisms, NULL);
	s_pModel->chooseSPNEGO(mechs);
	env->ReleaseStringChars(mechanisms, mechs);

	s_pModel->chooseCredentials(env);
	if (env->ExceptionCheck()) return;

	s_pModel->chooseContextRequirements(ISC_REQ_IDENTIFY);

}

JNIEXPORT jobjectArray JNICALL Java_net_za_darkskies_security_sasl_SSPIImpl_getClientMechanismNames
  (JNIEnv *env, jclass jobj, jobject map)
{
	return getMechanismNames(env, jobj, false, map);
}

JNIEXPORT void JNICALL Java_net_za_darkskies_security_sasl_SSPIImpl_serverInit
  (JNIEnv *env, jobject jobj, jstring mechanism, jstring protocol, jstring serverName, jobject props, jobject cbh)
{

	Model* s_pModel = new Model(env, true, _err);
	env->SetLongField(jobj, fidModel, (long long)s_pModel);

	const jchar *mech = env->GetStringChars(mechanism, NULL);

	s_pModel->chooseSpecificSSP(mech);

	env->ReleaseStringChars(mechanism, mech);

	s_pModel->chooseCredentials(env);
	if (env->ExceptionCheck()) return;

}

JNIEXPORT jobjectArray JNICALL Java_net_za_darkskies_security_sasl_SSPIImpl_getServerMechanismNames
  (JNIEnv *env, jclass jobj, jobject map)
{
	return getMechanismNames(env, jobj, true, map);
}

jobjectArray getMechanismNames
  (JNIEnv *env, jclass jobj, bool isServer, jobject map)
{
	Model* s_pModel = new Model(env, isServer, _err);

	ULONG cSSPs;
	PSecPkgInfo pSSPs;
	s_pModel->enumSSPs(env, &cSSPs, &pSSPs);
	if (env->ExceptionCheck()) {
		s_pModel->freeContextBuffer(pSSPs);
		return NULL;
	}

	int cMechs = 0;
	for (ULONG i = 0; i < cSSPs; ++i) {
		if (pSSPs[i].fCapabilities & SECPKG_FLAG_CONNECTION && 
			pSSPs[i].fCapabilities & SECPKG_FLAG_TOKEN_ONLY && 
			pSSPs[i].wRPCID != RPC_C_AUTHN_GSS_NEGOTIATE)
		{
			cMechs++;
		}
#ifdef _DEBUG
		wprintf(L"[%i] %08x, %i - %s\n", cMechs, pSSPs[i].fCapabilities, pSSPs[i].wRPCID, pSSPs[i].Name);
#endif
	}
	jobjectArray mechNames = env->NewObjectArray(cMechs,
		env->FindClass("java/lang/String"),
		NULL);

	cMechs = 0;
	for (ULONG i = 0; i < cSSPs; ++i) {
		if (pSSPs[i].fCapabilities & SECPKG_FLAG_CONNECTION && 
			pSSPs[i].fCapabilities & SECPKG_FLAG_TOKEN_ONLY && 
			pSSPs[i].wRPCID != RPC_C_AUTHN_GSS_NEGOTIATE)
			env->SetObjectArrayElement(
			mechNames,cMechs++,env->NewString(pSSPs[i].Name, (jsize)wcslen(pSSPs[i].Name)));
	}

	s_pModel->freeContextBuffer(pSSPs);

	delete s_pModel;

	return mechNames;
}

// Client
JNIEXPORT jbyteArray JNICALL Java_net_za_darkskies_security_sasl_SSPIImpl_evaluateChallenge
  (JNIEnv *env, jobject jobj, jbyteArray challenge)
{
	Model* s_pModel = (Model*)env->GetLongField(jobj, fidModel);

	const int bufsize = 4096 * 4;
	BYTE	s_msgBuf[bufsize];	// a big buffer eliminates buffer management worries
	DWORD	s_cbMsgBuf;			// length of message + security trailer (signature/encryption info)

	jsize size = env->GetArrayLength(challenge);
	jbyte *abyte = env->GetByteArrayElements(challenge, false);

	// psuedo-scope for i
	{
#ifdef _DEBUG
		wprintf(L"evaluateChallenge in: ");
#endif
		int i;
		for (i = 0; i < size && i < bufsize; i++) {
			s_msgBuf[i] = abyte[i];
#ifdef _DEBUG
			if (i % 0x10 == 0) { wprintf(L"\n%08x  ", i); }
			if (i % 0x10 == 8) { wprintf(L" "); }
			wprintf(L"%02x ", s_msgBuf[i]);
#endif
		}
#ifdef _DEBUG
		wprintf(L"\n%08x\n", i);
#endif
	}
	env->ReleaseByteArrayElements(challenge,abyte,0);
	s_cbMsgBuf = size;

	void* pOutput;
	DWORD cbOutput = 0;
	bool bContinueNeeded;
	ULONG grfCtxAttrs;

	wchar_t szSPN[256];

	jboolean iscopy;
	jstring authorizationID = (jstring)env->GetObjectField(jobj, fidAuthorizationID);
	const jchar *spn = env->GetStringChars(authorizationID, &iscopy);
	wcsncpy_s(szSPN, 256, spn, _TRUNCATE);
	env->ReleaseStringChars(authorizationID, spn);

	s_pModel->initializeSecurityContext(env, szSPN,
		s_msgBuf, s_cbMsgBuf,
		&pOutput, &cbOutput,
		&bContinueNeeded, &grfCtxAttrs);
	if (env->ExceptionCheck()) return NULL;

	jbyteArray response = env->NewByteArray(cbOutput);
	if (response) {
		abyte = env->GetByteArrayElements(response, false);
#ifdef _DEBUG
		wprintf(L"evaluateChallenge out: ");
#endif
		DWORD i;
		for (i = 0; i < cbOutput && i < bufsize; i++) {
			abyte[i] = ((byte*)pOutput)[i];
#ifdef _DEBUG
			if (i % 0x10 == 0) { wprintf(L"\n%08x  ", i); }
			if (i % 0x10 == 8) { wprintf(L" "); }
			wprintf(L"%02x ", ((byte*)pOutput)[i]);
#endif
		}
#ifdef _DEBUG
		wprintf(L"\n%08x\n", i);
#endif
		env->ReleaseByteArrayElements(response,abyte,0);
	}

	if (!bContinueNeeded) {
		env->SetBooleanField(jobj, fidCompleted, 1);
	}

	return response;
}

// Server
JNIEXPORT jbyteArray JNICALL Java_net_za_darkskies_security_sasl_SSPIImpl_evaluateResponse
  (JNIEnv *env, jobject jobj, jbyteArray response)
{
	Model* s_pModel = (Model*)env->GetLongField(jobj, fidModel);

	const int bufsize = 4096 * 4;
	BYTE	s_msgBuf[bufsize];	// a big buffer eliminates buffer management worries
	DWORD	s_cbMsgBuf;			// length of message + security trailer (signature/encryption info)

	jsize size = env->GetArrayLength(response);
	jbyte *abyte = env->GetByteArrayElements(response, false);

	// psuedo-scope for i
	{
#ifdef _DEBUG
		wprintf(L"evaluateResponse in: ");
#endif
		int i;
		for (i = 0; i < size && i < bufsize; i++) {
			s_msgBuf[i] = abyte[i];
#ifdef _DEBUG
			if (i % 0x10 == 0) { wprintf(L"\n%08x  ", i); }
			if (i % 0x10 == 8) { wprintf(L" ", 0); }
			wprintf(L"%02x ", s_msgBuf[i]);
#endif
		}
#ifdef _DEBUG
		wprintf(L"\n%08x\n", i);
#endif
	}
	env->ReleaseByteArrayElements(response,abyte,0);
	s_cbMsgBuf = size;

	void* pOutput;
	DWORD cbOutput = 0;
	bool bContinueNeeded;
	ULONG grfCtxAttrs;

#ifdef _DEBUG
	wprintf(L"Trying acceptSecurityContext...\n");
#endif
	s_pModel->acceptSecurityContext(env, 
		s_msgBuf, s_cbMsgBuf,
		&pOutput, &cbOutput,
		&bContinueNeeded, &grfCtxAttrs);
#ifdef _DEBUG
	wprintf(L"acceptSecurityContext back, checking for exception...\n");
#endif
	if (env->ExceptionCheck()) return NULL;
#ifdef _DEBUG
	wprintf(L"No exception...\n");
#endif
	
	jbyteArray challenge = env->NewByteArray(cbOutput);
	if (challenge) {
		abyte = env->GetByteArrayElements(challenge, false);
#ifdef _DEBUG
		wprintf(L"evaluateResponse out:", 0);
#endif
		DWORD i;
		for (i = 0; i < cbOutput && i < bufsize; i++) {
		abyte[i] = ((byte*)pOutput)[i];
#ifdef _DEBUG
			if (i % 0x10 == 0) { wprintf(L"\n%08x  ", i); }
			if (i % 0x10 == 8) { wprintf(L" ", 0); }
			wprintf(L"%02x ", ((byte*)pOutput)[i]);
#endif
		}
#ifdef _DEBUG
		wprintf(L"\n%08x\n", i);
#endif
		env->ReleaseByteArrayElements(challenge,abyte,0);
	}

	if (!bContinueNeeded) {
		wchar_t *pvSpn;
		s_pModel->queryContextAttributes(env, SECPKG_ATTR_NAMES, &pvSpn);
		if (env->ExceptionCheck()) {
			s_pModel->freeContextBuffer(pvSpn);
			return NULL;
		}
#ifdef _DEBUG
		wprintf(L"pvSpn:  %s\n", pvSpn);
#endif
		wchar_t *pvName;
		pvName = wcschr(pvSpn, '\\');
		if (pvName != NULL) {
			*pvName=0;
			pvName++;
		} else {
			pvName = pvSpn;
		}
#ifdef _DEBUG
		wprintf(L"pvSpn:  %s\n", pvSpn);
		wprintf(L"pvName: %s\n", pvName);
#endif
		jstring realm = env->NewString(pvSpn, (jsize)wcslen(pvSpn));
		jstring name = env->NewString(pvName, (jsize)wcslen(pvName));
		s_pModel->freeContextBuffer(pvSpn);

		env->CallVoidMethod(jobj, midCompleteAuthorization, realm, name);
	}

	return challenge;
}

JNIEXPORT void JNICALL Java_net_za_darkskies_security_sasl_SSPIImpl_finalize
  (JNIEnv *env, jobject jobj)
{
#ifdef _DEBUG
	wprintf(L"finalize\n");
#endif
	Model* s_pModel = (Model*)env->GetLongField(jobj, fidModel);
	delete s_pModel;

	env->SetLongField(jobj, fidModel, 0);
}
